{-# LANGUAGE BangPatterns, TupleSections, NoMonomorphismRestriction, CPP #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Dpfs.DavisPutnamFiniteSets
-- Copyright   :  (c) Thomas Ilgner 2012
-- Stability   :  provisional
-- Portability :  portable
--
-- Implementation of two extensions of the DPLL algorithm (Davis-Putnam-Logemann-Loveland)
--
-- [@First Extension:@] 
--
-- [@Second Extension:@] 
-----------------------------------------------------------------------------
module Dpfs.DavisPutnamFiniteSets (
	   
	    -- ** Datatype
	    Atom(Xin,Calc)  -- instance Eq,Show
	    
	    -- * Davis-Putnam procedures
	    ,dpfs
	    ,dpfsAll
	    ,dpfsSat
	    ,dpfsSatAllSolutions
	    
	    -- * Rules: 1st Extension
	    ,findTrueFalse
	    ,resolveTrueFalse
	    ,findSimilarAtomUnits
	    ,substituteSimilarUnits
	    ,findUnit'
	    ,resolveUnit
	    ,findPureLiteral
	    ,resolvePureLiteral
	    
	    -- * Rules: 2nd Extension
	    ,findUnit'
	    ,resolveUnitCalculus
	    ,findSimpRule3
	    ,resolveSimpRule3
	    ,findFalse_Calculus
	    ,resolveFalse_Calculus
	    ,findSimpUnitCalculus
	    ,resolveCalculus1
	    ,findSelection2
	    ,setSplit2
	    
	    -- * Utilities
	    ,lookup'
	    ,setToInt
	    ,dpcnf
	    ,expectJust
	    ,toList'
	    ) where

import Debug.Trace
import Data.List
import Dpfs.Simp
import Dpfs.Parser
import Data.Char
import qualified Data.IntSet as Set
import qualified Data.IntMap.Strict as IntMap
import qualified Data.Map.Strict as Map
import Data.StrictPair
import Data.Maybe

{--------------------------------------------------------------------
  Types
--------------------------------------------------------------------}
data Atom = Xin  {-# UNPACK #-} !Int !Set.IntSet
	  | Calc !Set.IntSet !Set.IntSet
	  deriving (Eq,Show)
            
type Klausel = [Atom]
type KlauselMenge = [Klausel]
type AtomTupel = (Int, Set.IntSet)
type FiniteSet = Int
type CalculusSwitch = Int
type SolutionSet = IntMap.IntMap Set.IntSet

-- | 'dpfs' function executes the following parts:
-- 
-- * parsing the formula 
--
-- * cnf transformation
--
-- * dp procedure extension 1 
dpfs :: FiniteSet -> [Char] -> [Char]
dpfs fset expr = let (varSubs, expr1) = numPropSubs fset . parseProp $ expr
                     cnf'      = cnf fset expr1
                     result    = dpfsSat 0 fset (dpcnf fset $ cnf')
                     resWVars  = backSubs result varSubs
                     resWVarsS = map (\(x,x1) -> x ++ "=" ++ show (Set.toList x1)) resWVars                     
                 in 
                     if   cnf' == (Pand [])
                      then  "Formel ist eine Tautologie"
                     else
                      if null resWVars  
                       then "Formel ist ein Widerspruch"
		       else "Modell: " ++ (foldr1 (\x y-> x ++ (", " ++ y)) resWVarsS)

-- | 'dpfsAll' function executes the following parts:
-- 
-- * parsing the formula 
--
-- * cnf transformation
--
-- * dp procedure extension 1 and tries to find all solutions 
dpfsAll :: Int -> [Char] -> [[(String, Set.IntSet)]]
dpfsAll fset expr = let (varSubs, expr1) = numPropSubs fset . parseProp $ expr
                        cnf' = cnf fset expr1
                        ress = dpfsSatAllSolutions 0 fset (dpcnf fset cnf')
                        ares = map (\x-> backSubsAll x varSubs) ress
		      in
			ares

{-
-}
dpcnf :: FiniteSet -> Pexpr (Int,[Int]) -> KlauselMenge
dpcnf fset (Pand xs) =  
  let res = if xs == [Por [Ptrue]] then   [] 
            else if xs == [Pfalse] || xs == [Por [Pfalse]]  then [[]]
             else map (dpcnfor fset) xs 
   in res     
 
dpcnfor :: Int -> Pexpr (Int, [Int]) -> Klausel
dpcnfor fset (Por xs) = map (dpcnfint fset) xs

dpcnfint :: Int -> Pexpr (Int, [Int]) -> Atom
dpcnfint _    (Pvar (x,s)) = Xin x (Set.fromList s)
dpcnfint fset (Pnot  (Pvar (x,s))) = Xin x ((Set.fromList [1..fset]) `Set.difference` (Set.fromList s) )
dpcnfint _ x  = error ("Error while converting cnf to clause set " ++ show x)  

backSubsAll :: Klausel -> [(String, Int)] -> [(String, Set.IntSet)]
backSubsAll [] _ = []
backSubsAll (x:xs) varSubs = backSubsAll' x varSubs : backSubsAll xs varSubs
  where backSubsAll' lit@(Xin x set) ((var,num):rest) = 
	  if x == num then (var,set)
          else backSubsAll' lit rest

backSubs :: SolutionSet -> [(String, Int)] -> [(String, Set.IntSet)]          
backSubs m varSubs = IntMap.foldlWithKey' (backSubs' varSubs) [] m
  where backSubs' ((var,num):rest) l k v =
	  if k == num then (var,v) : l
	  else backSubs' rest l k v 

{--------------------------------------------------------------------
  Davis-Putnam Procedures
--------------------------------------------------------------------}

{- | 'dpfsSat' function describes the procedure of the Davis-Putnam algorithm and returns one solutions if one exists.
-}
dpfsSat :: CalculusSwitch -> FiniteSet -> KlauselMenge -> SolutionSet
dpfsSat 1 fset klauselMenge = dpfsSat' 1 fset klauselMenge IntMap.empty
dpfsSat 0 fset klauselMenge = dpfsSat' 0 fset klauselMenge IntMap.empty

dpfsSat' :: CalculusSwitch -> FiniteSet -> KlauselMenge -> SolutionSet -> SolutionSet
dpfsSat' 0 _ [] l = l
dpfsSat' 0 f !k !l
  | [] `elemStrict` k = IntMap.empty
  | isJust (fTF) = dpfsSat' 0 f (resolveTrueFalse f (fromJust fTF) k) l
  | isJust (fSU) = dpfsSat' 0 f ((substituteSimilarUnits (fromJust fSU) k)) l
  | isJust (fU)  = let !(u1,u2) = fromJust fU in dpfsSat' 0 f ((resolveUnit (u1,u2) k )) (IntMap.insertWith (Set.intersection) u1 u2 l) 
  | isJust (fP)  = let !(u1,u2) = fromJust fP in dpfsSat' 0 f ((resolvePureLiteral (u1,u2) k)) (IntMap.insertWith (Set.intersection) u1 u2 l)
  | otherwise = let (m1,m2) = findBestLiteral k
		    (k1,k2) = strictPair m1 (Set.difference (Set.fromList [1..f]) m2)
		    newMap  = (IntMap.insertWith (Set.intersection) m1 m2 l)
		    d = let k' = (resolveUnit (m1,m2) k) in k' `seq` (dpfsSat' 0 f k' newMap)
		in 
		  if IntMap.null d
		    then
		      dpfsSat' 0 f (resolveUnit (k1,k2) k) (IntMap.insertWith (Set.intersection) k1 k2 l)
		    else
		      d
  where
    fTF = findTrueFalse f k
    fSU = findSimilarAtomUnits f k
    fU  = findUnit' k
    fP  = findPureLiteral k
dpfsSat' 1 _ [] l =  l
dpfsSat' 1 !f !k l
  | [] `elemStrict` k = IntMap.empty
  | isJust fU  = dpfsSat' 1 f ( resolveUnitCalculus v1 k ) (IntMap.insertWith (Set.intersection) (fst v1) (snd v1) l)
  | isJust fSR = let l' = resolveSimpRule3 v4 k l in dpfsSat' 1 f k l'
  | isJust fF  = dpfsSat' 1 f ( resolveFalse_Calculus k ) l
  | isJust fSU = if IntMap.null (snd v3) then dpfsSat' 1 f ( resolveCalculus1 (fst v3) k l) l else dpfsSat' 1 f ( resolveCalculus1 (fst v3) k (snd v3)) (snd v3)
  | isJust fSe = dpfsSat' 1 f (setSplit2 v2 k l) l
  | otherwise  = fSplit minUnit
  where
    fU  = findUnit' k
    fF  = findFalse_Calculus k
    fUM = findUnitMultiple $ lToList2 l
    fSU = findSimpUnitCalculus k l
    fSR = findSimpRule3 k l
    fSe = findSelection2 k l
    v1  = expectJust fU
    v2  = expectJust fSe
    v3  = expectJust fSU
    v4  = expectJust fSR
    minUnit = enumerate_BestLiteral (snd $ findBestLiteral_Calculus k)
    fSplit [] = IntMap.empty
    fSplit (x:xs)
      | IntMap.null expr = fSplit xs
      | otherwise = expr
      where expr = (dpfsSat' 1 f k (IntMap.insertWith (Set.intersection) (fst x) (snd x) l))

{- | 'dpfsAllSolutions' function describes the procedure of the Davis-Putnam algorithm and returns all solutions if any exists.
-}
dpfsSatAllSolutions :: CalculusSwitch -> FiniteSet -> KlauselMenge -> KlauselMenge
dpfsSatAllSolutions 0 fset klauselMenge = dpfsSatAllSolutions' 0 fset klauselMenge [[]]
dpfsSatAllSolutions 1 fset klauselMenge = dpfsSatAllSolutions' 1 fset klauselMenge [[]]

dpfsSatAllSolutions' :: CalculusSwitch -> FiniteSet -> KlauselMenge -> KlauselMenge -> KlauselMenge
dpfsSatAllSolutions' 0 _ [] l = resolveDuplicateLiteralsAll l
dpfsSatAllSolutions' 0 f !k l
  | [] `elemStrict` k = []
  | isJust (fTF) = dpfsSatAllSolutions' 0 f (resolveTrueFalse f (fromJust fTF) k) l
  | isJust (fSU) = dpfsSatAllSolutions' 0 f ((substituteSimilarUnits (fromJust fSU) k)) l
  | isJust (fU)  = let !(u1,u2) = fromJust fU in dpfsSatAllSolutions' 0 f ((resolveUnit (u1,u2) k )) ( map ((Xin u1 u2):) l)
  | isJust (fP)  = let !(u1,u2) = fromJust fP in dpfsSatAllSolutions' 0 f ((resolvePureLiteral (u1,u2) k)) ( map ((Xin u1 u2):)l)
  | otherwise = let !(m1,m2) = findBestLiteral k
		    !(k1,k2)  = (m1, (Set.difference (Set.fromList [1..f]) m2))
		in 
		  (dpfsSatAllSolutions' 0 f (resolveUnit (m1,m2) k) ( map ((Xin m1 m2):) l)) ++ 
		  (dpfsSatAllSolutions' 0 f (resolveUnit (k1,k2) k) ( map ((Xin k1 k2):) l))
  where
    fTF = findTrueFalse f k
    fSU = findSimilarAtomUnits f k
    fU  = findUnit' k
    fP  = findPureLiteral k
{- 
---------------------------------------------------------------------------------------------------------------------------------------------------------------------------
   findTrueFalse
--------------------------------------------------------------------------------------------------------------------------------------------------------------------------- 
-}

{- | 'findTrueFalse' runs through the clause set and returns an atom which can be simplify.
-}
findTrueFalse :: Int -> KlauselMenge -> Maybe Atom 
findTrueFalse _ [] = Nothing
findTrueFalse fset ([Xin !a !b]:xs)
  | Set.size b == fset = Just (Xin a b)
  | Set.null b      = Just (Xin a b)
  | otherwise = findTrueFalse fset xs
findTrueFalse fset (x:xs) = findTrueFalse' fset x xs
  where
    findTrueFalse' :: Int -> Klausel -> KlauselMenge -> Maybe Atom
    findTrueFalse' s [] rest = findTrueFalse s rest					
    findTrueFalse' s ((Xin !a !b):ks) rest
      | Set.size b == s = Just (Xin a b)
      | Set.null b      = Just (Xin a b)
      | otherwise = findTrueFalse' s ks rest
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE findTrueFalse #-}
#else
{-# INLINE findTrueFalse #-}
#endif
  
{- | 'resolveTrueFalse' simplfies the clause set for the given atom.
-}
resolveTrueFalse :: Int -> Atom -> KlauselMenge -> KlauselMenge
resolveTrueFalse _ _ [] = []
resolveTrueFalse fset unit@(Xin !a !b) !(k:ks)
    | fset `seq` unit `seq` k `seq` ks `seq` False = undefined
    | cmp && (b == Set.fromList [1..fset]) = resolveTrueFalse fset unit ks -- clausel gesamt true
    | cmp && (length k == 1) && fclause           = [] : resolveTrueFalse fset unit ks
    | cmp                                         = resolveTrueFalse' fset unit k ks
    | otherwise                                   = k : resolveTrueFalse fset unit ks
    where cmp = (a,b) `elem''` k
          fclause = Set.null b
          resolveTrueFalse' :: Int -> Atom -> Klausel -> KlauselMenge -> KlauselMenge
          resolveTrueFalse' fset unit@(Xin !a !b) [] ks = resolveTrueFalse fset unit ks   
          resolveTrueFalse' fset unit@(Xin !a !b) !(x:xs) ks
	    | (unit) == x = resolveTrueFalse' fset unit xs ks
	    | otherwise  = [x] : resolveTrueFalse' fset unit xs ks                                 
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE resolveTrueFalse #-}
#else
{-# INLINE resolveTrueFalse #-}
#endif
{- 
---------------------------------------------------------------------------------------------------------------------------------------------------------------------------
   findDuplicateLiterals
--------------------------------------------------------------------------------------------------------------------------------------------------------------------------- 
-}
    
findDuplicateLiterals :: KlauselMenge -> Maybe [Atom]
findDuplicateLiterals [] = Nothing
findDuplicateLiterals (x:xs)
    | length x >= 2 = findDuplicateLiterals' x xs
    | otherwise    = findDuplicateLiterals xs
  where
    findDuplicateLiterals' :: Klausel -> KlauselMenge -> Maybe Klausel    
    findDuplicateLiterals' x xs = let dup = nubBy (\(Xin a b) (Xin a1 b1) -> a == a1 ) x
				  in
				  case (length dup) /= (length x) of
				    True  -> Just x
				    False -> findDuplicateLiterals xs
                               
resolveDuplicateLiterals :: Klausel -> Klausel                                
resolveDuplicateLiterals x = nubBy (\(Xin a b) (Xin a1 b1) -> a == a1 ) x

resolveDuplicateLiteralsAll :: KlauselMenge -> KlauselMenge 
resolveDuplicateLiteralsAll []     = []                                     
resolveDuplicateLiteralsAll (x:xs) = (nubBy (\(Xin a b) (Xin a1 b1) -> a == a1 ) x) :  resolveDuplicateLiteralsAll xs                               

{- 
---------------------------------------------------------------------------------------------------------------------------------------------------------------------------
   findUnit & resolveUnit
--------------------------------------------------------------------------------------------------------------------------------------------------------------------------- 
-}

{- | The 'findUnit'' function return the first unit (a elem M') from the clause set if one exist.
-}
findUnit' :: KlauselMenge -> Maybe AtomTupel
findUnit' [] = Nothing
findUnit' ([Xin x y]:_) = Just (x `strictPair` y)
findUnit' ([Calc !x !y]:xs)
  | ((Set.size x) == 1) && ((Set.size y) == 1) =  Just ((head (Set.toList x)),y)
  | otherwise = findUnit' xs
findUnit' (_:xs) = findUnit' xs

{- | The 'findUnitMultiple' function return the first unit (a elem M') from the clause set,  if one exist.
-}
findUnitMultiple :: Klausel -> Maybe AtomTupel
findUnitMultiple [] = Nothing
findUnitMultiple ((Xin x y):xs)
  | (Set.size y) > 1 = Just (x `strictPair` y)
  | otherwise = findUnitMultiple xs
findUnitMultiple (_:xs) = findUnitMultiple xs
  
resolveUnit :: AtomTupel -> KlauselMenge -> KlauselMenge    
resolveUnit unit@(u1,u2) xs = foldl' (f1 u1) [] xs
  where f1 x1 lst x
	  | not (isRemoveable unit x) = lst
	  | x1 `elem'` x = resolveUnit' unit x : lst
	  | otherwise = x : lst
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE resolveUnit #-}
#else
{-# INLINE resolveUnit #-}
#endif
	  
resolveUnit' :: AtomTupel -> Klausel -> Klausel   
resolveUnit' unit xs = foldl' (f' unit) [] xs
 where f' (!x1,!y1) lst x
        | (x1 == (showLiteral x) && ((showSet x) `Set.isProperSubsetOf` y1) && y1 /= (showSet x)) = (Xin x1 (showSet x)) : lst
        | x1 == (showLiteral x)   = (Xin x1 (Set.intersection y1 (showSet x))) : lst
        | otherwise  = (Xin (showLiteral x) (showSet x)) : lst
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE resolveUnit' #-}
#else
{-# INLINE resolveUnit' #-}
#endif

{- |The 'removeable' function indicates whenever an given clause could be removed from the clause set.
-}
isRemoveable :: (Int,Set.IntSet) -> Klausel -> Bool
isRemoveable        _ []                                         = True
isRemoveable (!x1,!y1) ((Xin !x2 !y2):ks) 
    |  (x1 == x2 && y1 `Set.isSubsetOf` y2) = False
    |  (x1 == x2 && y2 `Set.isProperSubsetOf` y1 && y1 /= y2) = True
    |  otherwise = isRemoveable (x1,y1) ks
isRemoveable (!x1,!y1) (_:ks) = isRemoveable (x1,y1) ks
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE isRemoveable #-}
#else
{-# INLINE isRemoveable #-}
#endif

{- 
---------------------------------------------------------------------------------------------------------------------------------------------------------------------------
   findSimilarAtomUnits
--------------------------------------------------------------------------------------------------------------------------------------------------------------------------- 
-}

{- |The 'findSimilarAtomUnits' function interates through the clause set and returns the first unit a.
-}
findSimilarAtomUnits :: Int -> KlauselMenge -> Maybe Atom
findSimilarAtomUnits fset [] = Nothing
findSimilarAtomUnits fset w@([Xin x y]:xs) = findSimilarAtomUnits' fset (x,y) xs
  where
    findSimilarAtomUnits' fset (a,b) lst = let x = foldr (\x l -> if ((showLiteral (head x)) == a) then (head x) : l else l ) [] $ (filter (\x -> length x == 1) lst)
				    in
				      case x of
					[] -> Nothing
					xs -> Just (Xin a (resolveSimilarAtomUnits' fset (xs ++ [Xin a b]) ))
findSimilarAtomUnits fset (_:xs) = findSimilarAtomUnits fset xs
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE findSimilarAtomUnits #-}
#else
{-# INLINE findSimilarAtomUnits #-}
#endif

resolveSimilarAtomUnits' :: Int -> Klausel -> Set.IntSet
resolveSimilarAtomUnits' fset !xs = foldl' (Set.intersection) (Set.fromList [1..fset]) (map (\(Xin a b) -> b) xs)

resolveSimilarAtomUnits :: Klausel -> Klausel -> Atom
resolveSimilarAtomUnits [] ys = head ys
resolveSimilarAtomUnits !((Xin x1 x2):xs) [] = resolveSimilarAtomUnits xs [Xin x1 x2]
resolveSimilarAtomUnits !((Xin x1 x2):xs) [Xin y1 y2] = resolveSimilarAtomUnits xs [Xin y1 (Set.intersection x2 y2)]

substituteSimilarUnits :: Atom -> KlauselMenge -> KlauselMenge
substituteSimilarUnits unit xs = [[unit]] ++ (foldr (\x lst -> substituteSimilarUnits' unit x lst) [] xs)
  where    
    substituteSimilarUnits' unit ([Xin x y]) lst
      | (showLiteral unit) /= x = [Xin x y] : lst
      | otherwise = lst
    substituteSimilarUnits' unit ([x]) lst = [x] : lst
    substituteSimilarUnits' unit     x lst = x : lst
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE substituteSimilarUnits #-}
#else
{-# INLINE substituteSimilarUnits #-}
#endif
    
{- | deprecated function
-}
substituteSimilarUnitsO :: AtomTupel -> KlauselMenge -> KlauselMenge
substituteSimilarUnitsO unit xs = [[Xin (fst unit) (snd unit)]] ++ (foldr (\x lst -> substituteSimilarUnits' unit x lst) [] xs)
  where    
    substituteSimilarUnits' unit ([Xin x y]) lst
      | (fst unit) /= x = [Xin x y] : lst
      | otherwise = lst
    substituteSimilarUnits' unit ([x]) lst = [x] : lst
    substituteSimilarUnits' unit     x lst = x : lst    

    
{- 
---------------------------------------------------------------------------------------------------------------------------------------------------------------------------
   FindPureLiterals
--------------------------------------------------------------------------------------------------------------------------------------------------------------------------- 
-}
    
{- | Add documentation here
-}
filterList :: Atom -> Klausel -> Bool
filterList _ [] = False
filterList x@(Xin x1 y1) ((Xin x2 y2):xs) = (x1 == x2) || filterList x xs
filterList x@(Xin x1 y1) (_:xs) = filterList x xs
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE filterList #-}
#else
{-# INLINE filterList #-}
#endif

{- | Add documentation here
-}
collectLit :: AtomTupel -> KlauselMenge -> AtomTupel
collectLit literal [] = literal
collectLit literal (x:xs) = collectLit' literal x xs
  where
    collectLit' (x1,y1) [] [] = (x1,y1)
    collectLit' (x1,y1) [] klauselMenge = collectLit (x1,y1) klauselMenge
    collectLit' (x1,y1) ((Xin x2 y2):xs) ks
	| x1 == x2 = collectLit' (x1,(Set.intersection y1 y2)) xs ks
	| otherwise = collectLit' (x1,y1) xs ks
    collectLit' (x1,y1) (_:xs) ks =  collectLit' (x1,y1) xs ks
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE collectLit #-}
#else
{-# INLINE collectLit #-}
#endif

{- | The 'findPureLiteral' function returns the first pure literal.
-}
findPureLiteral :: KlauselMenge -> Maybe AtomTupel
findPureLiteral !(x:xs) = findPureLitR [] x xs
  where
    findPureLitR :: [Int] -> Klausel -> KlauselMenge -> Maybe AtomTupel
    findPureLitR _ [] [] = Nothing
    findPureLitR literals [] !(k:ks)
	= findPureLitR literals k ks
    findPureLitR literals !((Xin x y):ks) klauselMenge
	| x `elem` literals =  findPureLitR literals ks klauselMenge -- bereits gefundene Literale
	| otherwise =
	    let
		new    = filter (filterList (Xin x y)) klauselMenge
		newSet = snd(collectLit (x,y) new)
		    in              
			if Set.null newSet 
			    then findPureLitR (x:literals) ks klauselMenge
			    else Just (x,newSet)
    findPureLitR literals !(_:ks) klauselMenge = findPureLitR literals ks klauselMenge
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE findPureLiteral #-}
#else
{-# INLINE findPureLiteral #-}
#endif

{- |The 'resolvePureLiteral' builds up a new clause set while skipping clauses which contains the pure unit.
-}
resolvePureLiteral :: AtomTupel -> KlauselMenge -> KlauselMenge
resolvePureLiteral (x,y) [] = [[Xin x y]]
resolvePureLiteral (x,y) (k:ks)
    | x `elem'` k  = resolvePureLiteral (x,y) ks
    | otherwise = k : resolvePureLiteral (x,y) ks
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE resolvePureLiteral #-}
#else
{-# INLINE resolvePureLiteral #-}
#endif
    
{- 
-----------------------------------------------------------------------------------------------------------------------------------
  findBestLiteral
----------------------------------------------------------------------------------------------------------------------------------- 
-}

{- | The 'findBestLiteral' function returns an literal of the given claus set where the literal occours at most in any clauses and the finite set is possibly small.
-}
findBestLiteral :: KlauselMenge -> (IntMap.Key, Set.IntSet)				    
findBestLiteral k = let !k' = concat k
			!min = fst $ IntMap.foldlWithKey' (\(x1,x2) y z -> if x1 == 0 then (y,z) else if x2 > z then (y,z) else (x1,x2) ) (0,0) $ foldl' (\map1 (Xin k1 _) -> IntMap.insertWith (+) k1 1 map1) IntMap.empty k'
			(v1,v2) = findSmallestSet min k'
		     in
		      v1 `strictPair` v2
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE findBestLiteral #-}
#else
{-# INLINE findBestLiteral #-}
#endif

{- 
---------------------------------------------------------------------------------------------------------------------------------------------------------------------------
   Calculus Extension
--------------------------------------------------------------------------------------------------------------------------------------------------------------------------- 
-}

{- | The 'findFalse_Calculus' find the first calc atom with one or both sets are empty.
-}
findFalse_Calculus :: KlauselMenge -> Maybe Atom
findFalse_Calculus [] = Nothing
findFalse_Calculus ([Calc !x !y]:xs)
  | (Set.null x || Set.null y) = Just (Calc x y)
  | otherwise = findFalse_Calculus xs 
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE findFalse_Calculus #-}
#else
{-# INLINE findFalse_Calculus #-}
#endif

{- | The 'resolveFalse_Calculus' function runs through the clause set and removes all atoms with one or both sets are empty.
-}
resolveFalse_Calculus :: KlauselMenge -> KlauselMenge
resolveFalse_Calculus klauselMenge = foldl' fCalFal [] klauselMenge
  where
    fCalFal lst x@[Calc !a !b]
      | (Set.null a && Set.null b) =  [] : lst
      | otherwise = x : lst
    fCalFal lst x@[Xin !a !b]
      | Set.null b =  [] : lst
      | otherwise = x : lst
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE resolveFalse_Calculus #-}
#else
{-# INLINE resolveFalse_Calculus #-}
#endif
    
{- | Add documentation here
-}   
findSelection :: KlauselMenge -> Bool
findSelection [] = False 
findSelection km@([Calc !x !y]:xs)
  | not (null u) && (length u > 1) = True
  | otherwise = findSelection xs
  where
    u = let r = resolveDuplicateLiterals $ findSelection'' (concat (map (\(Calc !z _) -> Set.toList z) [Calc x y])) km in (trace ("Sel " ++ (show (x,r))) r)
findSelection (_:xs) = findSelection xs
	    
{- | The 'findSelection' function returns an calc atom, if and only if this calc atom is splitable into two pieces. 
-}
findSelection2 :: KlauselMenge -> SolutionSet -> Maybe Atom
findSelection2 klauselMenge loesung = findSelection2' (Set.fromList (lToList loesung)) klauselMenge 

findSelection2' :: Set.IntSet -> KlauselMenge -> Maybe Atom
findSelection2' _ [] = Nothing
findSelection2' xset cl@([Calc !x !y]:xs)
  | sizeConstrait && (cmpSize) > 1 && cmpSize < (Set.size x) = Just (Calc x y)
  | otherwise = findSelection2' xset xs
  where
    sizeConstrait = (Set.size x) > 1
    cmpSize = Set.size $ inters
    inters  = Set.intersection xset x
findSelection2' xset (_:xs) = findSelection2' xset xs
    
{- | The 'setSplit' function splits the passed calc atom into two ore more pieces.
-}
setSplit2 :: Atom -> KlauselMenge -> SolutionSet -> KlauselMenge
setSplit2 unit@(Calc x1 x2) klauselMenge loesung = foldl' (fsplit unit) [] klauselMenge
  where
    fsplit c1@(Calc u1 u2) l x@[Calc !a !b]
      | constraitId && (length u > 1) && constraitSize && card1 = let (!a',!b') = ((Set.difference a n1),(Set.difference b n2))
								      subset	= findSelection' (Set.toList n1) Set.empty set
								  in
								    --(trace ("neu " ++ (show (n1,n2))) [Calc a' b'] : [Calc n1 n2] : l)
								    case (Set.size n1 `elem` [2..4] && (not (null subset))) of
								      True  -> [Calc a' b'] : ((fsplitSubset (Calc n1 n2)) ++ l)
								      False -> [Calc a' b'] : [Calc n1 n2] : l
      | constraitId && (length u > 1) && (card <  (length u))   = [] : l
      | otherwise = x : l
      where
	constraitId = (c1 == (head x))
	set	    = map (\(x,y) -> (Xin x y) ) $ IntMap.toList loesung
	u 	    = findSelection' (concat (map (\(Calc !z !w) -> (Set.toList z)) x)) (showCalcr x) set
	constraitSize = (Set.size u1) > (length u)
	card = Set.size $ foldl' (Set.union) Set.empty (map (\(Xin x y) -> y ) u)
	card1 = (card == (length u))
	(n1,n2) = ((foldl' (Set.union) Set.empty ((map (\(Xin a b) -> Set.fromList [a])) u)), (foldl' (Set.union) Set.empty ((map (\(Xin a b) -> b)) u)))
    fsplit _ l x = x : l
    fsplitSubset (Calc n3 n4) = let !setXin = filter (\x -> isXinAtom x) (concat klauselMenge)
				    !set = Set.fromList $ map (\(Xin a b) -> a ) setXin
				    set1 = map (\(x,y) -> (Xin x y) ) $ IntMap.toList loesung
				in
				  fsplitSubset' (Set.toList n3) set1
	where
	  fsplitSubset' _ [] = []
	  fsplitSubset' calcList ((Xin !y1 !y2):ys)
	    | y1 `elem` calcList = [Calc (Set.fromList [y1]) y2] : fsplitSubset' calcList ys
	    | otherwise = fsplitSubset' calcList ys

findSelection' :: [Int] -> Set.IntSet -> Klausel -> Klausel
findSelection' _ _ [] = []
findSelection' !set !set1 ((Xin !x !y):xs)
  | x `elem` set && (Set.isSubsetOf y set1) && ((Set.size y) == 1) = (Xin x y) : findSelection' set set1 xs
  | otherwise = findSelection' set set1 xs
findSelection' !set _ ((Xin !x !y):xs)
  | x `elem` set && ((Set.size y) == 1) = (Xin x y) : findSelection' set Set.empty xs
  | otherwise = findSelection' set Set.empty xs  
findSelection' !set set1 (_:xs) = findSelection' set set1 xs
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE findSelection' #-}
#else
{-# INLINE findSelection' #-}
#endif
	    
findSelection'' :: [Int] -> KlauselMenge -> Klausel
findSelection'' _ [] = []
findSelection'' !set ([Xin !x !y]:xs)
  | x `elem` set = (Xin x y) : findSelection'' set xs
  | otherwise = findSelection'' set xs
findSelection'' !set (_:xs) = findSelection'' set xs

setSplit :: KlauselMenge -> KlauselMenge
setSplit klauselMenge = foldl' fSplit [] klauselMenge
  where    
    fSplit l x@[Calc !a !b]
      | not (null u) && (length u > 1) && (card == (length u)) = let (a',b') = ((Set.difference a n1),(Set.difference b n2)) in if ((Set.null a') && (Set.null b')) then l else [Calc a' b'] : [Calc n1 n2] : l
      | not (null u) && (length u > 1) && (card <  (length u)) = [] : l
      | otherwise = x : l
      where
	u    = resolveDuplicateLiterals $ findSelection'' (concat (map (\(Calc !z _) -> Set.toList z) x)) klauselMenge
	card = Set.size $ foldl' (Set.union) Set.empty (map (\(Xin x y) -> y ) u)
	(n1,n2) = ((foldl' (Set.union) Set.empty ((map (\(Xin a b) -> Set.fromList [a])) u)), (foldl' (Set.union) Set.empty ((map (\(Xin a b) -> b)) u)))
    fSplit l x = x : l

collectXinMultipleElements :: AtomTupel -> KlauselMenge -> Klausel
collectXinMultipleElements unit klauselMenge = foldl' (collectXinMultipleElements' unit) [] klauselMenge
  where
    collectXinMultipleElements' :: AtomTupel -> Klausel -> Klausel -> Klausel
    collectXinMultipleElements' (x1,y1) lst c@[Calc x y]
      | mem = [Xin x1 (Set.intersection y1 y)]
      | otherwise = lst
      where
	mem  = Set.member x1 x
    collectXinMultipleElements' (x1,y1) lst c = lst

{- | The 'resolveUnitCalculus' function removes the given atom from the clause set.
-}    
resolveUnitCalculus :: AtomTupel -> KlauselMenge -> KlauselMenge
resolveUnitCalculus unit@(x1,y1) klauselMenge  = foldl' (resolveUnitCalculus' unit) [] klauselMenge
  where
    resolveUnitCalculus' :: AtomTupel -> KlauselMenge -> Klausel -> KlauselMenge
    resolveUnitCalculus' (x1,y1) lst l@[Xin x y]
      | (x1 == x) && (y1 == y) = lst
      | otherwise = l : lst
    resolveUnitCalculus' _ lst calc@[Calc x y] 
      | (Set.size x) == 1 && (Set.size y) == 1 = lst
      | otherwise = calc : lst
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE resolveUnitCalculus #-}
#else
{-# INLINE resolveUnitCalculus #-}
#endif

{- | The 'findSimpRule3' function runs through the clause set and returns an Atom if any simplifications are possible for rule 3.
-}
findSimpRule3 :: KlauselMenge -> SolutionSet -> Maybe ((Int, Set.IntSet), Atom)
findSimpRule3 klauselMenge loesung = findSimpRule3' set
  where
    findSimpRule3' [] = Nothing
    findSimpRule3' ((Xin x y):xs) = findSimpRule3'' (x,y) klauselMenge xs
    findSimpRule3'' _ [] ys = findSimpRule3' ys
    findSimpRule3'' (u1,u2) ([Calc u v]:xs) ys
      | (Set.member u1 u) && ((Set.size u2) == 1) && (not (Set.isSubsetOf u2 v)) = Just ((u1,u2),(Calc u v)) -- Zu False evaluieren IN REGEL 2
      | (Set.member u1 u) && ((Set.size u2) > 1) && (not (Set.isSubsetOf u2 v)) = Just ((u1,u2),(Calc u v))
      | otherwise = findSimpRule3'' (u1,u2) xs ys
    set = lToList2 loesung
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE findSimpRule3 #-}
#else
{-# INLINE findSimpRule3 #-}
#endif    

{- | The 'resolveSimpRule3' function updates the solutionSet for the given calc atom
-}
resolveSimpRule3 :: ((Int, Set.IntSet), Atom) -> KlauselMenge -> SolutionSet -> SolutionSet
resolveSimpRule3 ((u1,u2),(Calc x y)) klauselMenge loesung = IntMap.insertWith (Set.intersection) u1 (Set.intersection u2 y) loesung

{- | The 'findSimpUnitCalculus' function runs through the clause set and returns an Atom if any simplifications are possible. (Rule 1 + 2)
-}
findSimpUnitCalculus :: KlauselMenge -> SolutionSet -> Maybe (Atom,SolutionSet)
findSimpUnitCalculus klauselMenge loesung = findSimpUnitCalculus' set klauselMenge
  where
    findSimpUnitCalculus' _ [] = Nothing
    findSimpUnitCalculus' xset cl@([Calc !x !y]:xs)
      | cmpSize == (Set.size x) = Just ((Calc x y), IntMap.empty) -- ganze atom == true     
      | cmpSize == 1 && (not mem) = let newL = IntMap.insert (head $ Set.toList inter) (Set.intersection y (fromJust newY)) loesung in Just ((Calc x y),newL)
      | cmpSize == 1 = Just ((Calc x y), IntMap.empty) -- regel 1
      | otherwise = findSimpUnitCalculus' xset xs
      where
	inter	= Set.intersection xset x
	cmpSize = Set.size $ inter
	mem
	  | isJust newY = let u = fromJust newY
			  in
			    if (Set.null u)
			    then False
			    else
			    (Set.member (head $ Set.toList u) y)
	newY
	  | not (Set.null inter) = IntMap.lookup (head $ Set.toList inter) loesung
	  | otherwise = Nothing
    findSimpUnitCalculus' xset (_:xs) = findSimpUnitCalculus' xset xs
    !set = Set.fromList $ lToList loesung
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE findSimpUnitCalculus #-}
#else
{-# INLINE findSimpUnitCalculus #-}
#endif 
    
{- | The 'resolveCalculus' function simplifies the clause set an returns the updated clause set (Rule 1 + 2)
-}    
resolveCalculus1 :: Atom -> KlauselMenge -> SolutionSet -> KlauselMenge
resolveCalculus1 unit@(Calc x1 y1) klauselMenge loesung = foldl' (resolveCalculus' unit) [] klauselMenge
  where
    resolveCalculus' :: Atom -> KlauselMenge -> Klausel -> KlauselMenge
    resolveCalculus' (Calc x1 y1) lst c@[Calc x y] 
      | cmp && (cmpSize == 1) && mem = let newX1 = (Set.delete (head $ Set.toList interS) x)
					   newY1 = (Set.delete (head $ Set.toList (fromJust newY)) y)
				       in
					 if (Set.size newX1) == (Set.size newY1)
					  then
					    [Calc newX1 newY1] : lst
					 else
					  lst
      | cmp && (cmpSize == 1) = [] : lst -- Regel 3
      | cmp && (cmpSize == (Set.size x)) = lst -- alle elemente in loesungsmenge
      | otherwise = c : lst
      where
	cmp  = (x1 == x) && (y1 == y)
	cmpSize = Set.size $ interS	     
	mem
	  | isJust newY = let u = fromJust newY
			  in
			    if (Set.null u)
			    then False
			    else
			    (Set.member (head $ Set.toList u) y1)
	newY
	  | not (Set.null interS) = IntMap.lookup (head $ Set.toList interS) loesung
	  | otherwise = Nothing	
	sset = Set.isProperSubsetOf y1 y
	interS = Set.intersection x1 (Set.fromList $ lToList loesung) 
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE resolveCalculus1 #-}
#else
{-# INLINE resolveCalculus1 #-}
#endif
	
{- | Add documentation here
-}    
resolveCalculus :: AtomTupel -> KlauselMenge -> KlauselMenge
resolveCalculus unit@(x1,y1) klauselMenge 
  | Set.size y1 > 1 = let !u = atomToTupel $ head $ collectXinMultipleElements unit klauselMenge
		      in
			foldl' (resolveCalculus'' unit u) [] klauselMenge
  | otherwise = foldl' (resolveCalculus' unit) [] klauselMenge
  where
    resolveCalculus' :: AtomTupel -> KlauselMenge -> Klausel -> KlauselMenge
    resolveCalculus' (x1,y1) lst c@[Calc x y] 
      | (Set.fromList [x1] == x) && (y1 == y) = lst
      | mem && not sset = [] : lst
      | mem = ([Calc (Set.delete x1 x) (Set.difference y y1)]) : lst
      | otherwise = c : lst
      where
	mem  = Set.member x1 x
	sset = Set.isProperSubsetOf y1 y
    resolveCalculus'' :: AtomTupel -> AtomTupel -> KlauselMenge -> Klausel -> KlauselMenge
    resolveCalculus'' (x1,y1) (u1,u2) lst l@[Xin x y]
      | (x1,y1) == (x,y) = [Xin u1 u2] : lst
      | otherwise = l : lst
    resolveCalculus'' _ _ lst c@[Calc x y] = c : lst
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE resolveCalculus #-}
#else
{-# INLINE resolveCalculus #-}
#endif

{- | The 'findBestLiteral_Calculus1' functions returns one Calc Atom with the smallest set
-}
findBestLiteral_Calculus1 :: KlauselMenge -> (Int,(Set.IntSet, Set.IntSet))
findBestLiteral_Calculus1 klauselMenge = let !c = (concat klauselMenge)
					 in
					  Map.findMin $ Map.fromList $ map (\(Calc a b) -> ((Set.size b),(a,b)) ) (filter (\x -> isCalcAtom x ) c)
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE findBestLiteral_Calculus #-}
#else
{-# INLINE findBestLiteral_Calculus #-}
#endif

{- | The 'findBestLiteral_Calculus' functions returns one Calc Atom with the smallest set
-}
findBestLiteral_Calculus :: KlauselMenge -> (Int,(Set.IntSet, Set.IntSet))
findBestLiteral_Calculus klauselMenge = let !cl = (concat klauselMenge)
					in
					  IntMap.findMin $ IntMap.fromList $ map (\(Calc a b) -> ((Set.size b),(a,b))) cl

{- | The 'enumerate_BestLiteral' function returns for the given calc atom the set for the case differentiation
-}
enumerate_BestLiteral :: (Set.IntSet, Set.IntSet) -> [(Int,Set.IntSet)]
enumerate_BestLiteral (!x,!y)
  | Set.null x = []
  | otherwise = let (i,_) = Set.deleteFindMin x
		    b     = Set.toList y
		in
		  (foldr (\z l -> (i,(Set.singleton z)) : l) [] b)
#if __GLASGOW_HASKELL__ >= 700
{-# INLINABLE enumerate_BestLiteral #-}
#else
{-# INLINE enumerate_BestLiteral #-}
#endif
		  
{- 
---------------------------------------------------------------------------------------------------------------------------------------------------------------------------
   Utilities
--------------------------------------------------------------------------------------------------------------------------------------------------------------------------- 
-}
lToList :: SolutionSet -> [Int]
lToList loesung
  | IntMap.null loesung = []
  | otherwise = map (\(x,y) -> x ) $ IntMap.toList loesung
  
lToList2 :: SolutionSet -> [Atom]
lToList2 loesung
  | IntMap.null loesung = []
  | otherwise = map (\(x,y) -> Xin x y ) $ IntMap.toList loesung  

isXinAtom :: Atom -> Bool
isXinAtom (Xin _ _) = True
isXinAtom _ = False

isXinAtom' :: [Atom] -> Bool
isXinAtom' [Xin _ _] = True
isXinAtom' _ = False

isCalcAtom :: Atom -> Bool
isCalcAtom (Calc _ _) = True
isCalcAtom _	      = False

isCalcAtom' :: [Atom] -> Bool
isCalcAtom' [Calc _ _] = True
isCalcAtom' _	      = False

showCalcl :: [Atom] -> Set.IntSet
showCalcl [Calc a _] = a

showCalcr :: [Atom] -> Set.IntSet
showCalcr [Calc _ b] = b

showLiteral :: Atom -> Int
showLiteral (Xin x _) = x

showSet :: Atom -> Set.IntSet
showSet (Xin _ y) = y

showXin :: [Atom] -> Int
showXin [Xin x _] = x
showXin _ = 0

-- |The helper 'atomToTupel' returns an tupel from the Xin data constructor.
atomToTupel :: Atom -> (Int,Set.IntSet)
atomToTupel (Xin !x !y) = (x,y)

elemStrict :: (Eq a) => a -> [a] -> Bool
elemStrict _ []       = False
elemStrict !x (!y:ys)   = x==y || elemStrict x ys

-- |The helper 'elem'' indicates if the atom member of one given clause.
elem' :: Int -> [Atom] -> Bool
elem' _ []       = False
elem' x ((Xin x2 _):ys)   = x == x2 || elem' x ys
elem' x (_:ys) = elem' x ys

-- |The helper 'elem'' indicates if the atom member of one given clause.
-- elem1 :: Int -> [Atom] -> Bool
elem1 :: Int -> Klausel -> [Atom]
elem1 _ []       = []
elem1 fset (x@(Xin _ y2):ys)   = if ((Set.null y2) || (y2 == Set.fromList [1..fset])) then [x] else elem1 fset ys
elem1 fset (_:ys) = elem1 fset ys

-- |The helper 'elem'' indicates if the atom member of one given clause.
elem'' :: (Int, Set.IntSet) -> [Atom] -> Bool
elem'' _ []       = False
elem'' (x,y) ((Xin x2 y2):ys)   = (x == x2 && y == y2) || elem'' (x,y) ys

-- |The helper 'lookup'' function looks up a set of one literal a
lookup'                  :: Int -> [Atom] -> Maybe (Set.IntSet)
lookup' _key []         =  Nothing
lookup'  key ((Xin x y):xys)
    | key == x          =  Just y
    | otherwise         =  lookup' key xys

setToInt :: Int -> SolutionSet -> Int
setToInt key set = case IntMap.lookup key set of
                Just x -> head $ Set.toList x
                
toList' :: Set.IntSet -> [Int]
toList' t
  = Set.foldl' (\l x -> x : l) [] t

expectJust :: Maybe a -> a
expectJust (Just x) = x
expectJust Nothing  = error "Unexpected Nothing"

findSmallestSet :: Int -> [Atom] -> (IntMap.Key, Set.IntSet)
findSmallestSet k = IntMap.findMin . foldl' (\map1 (Xin k1 k2) -> if k == k1 then insertIfNew k1 k2 map1 else map1 ) IntMap.empty  		   

insertIfNew :: IntMap.Key -> Set.IntSet -> IntMap.IntMap Set.IntSet -> IntMap.IntMap Set.IntSet
insertIfNew k v m = case (IntMap.lookup k m) of
		   Just s -> if ((Set.size v) <= (Set.size s)) then IntMap.insert k v m else m
		   Nothing -> IntMap.insert k v m