{
module Dpfs.Parser where

import Data.Char
import qualified Control.OldException(evaluate,try,errorCalls)
import System.IO.Unsafe
}

%name prop Expr
%tokentype { Token }

%token
    selem  { (TokSElem $$, _, _) }
    all    { (TokAll,    _, _) }
    exists { (TokExists, _, _) }
    lpar   { (TokParL,   _, _) }
    rpar   { (TokParR,   _, _) }    	
    ':'    { (TokColon,  _, _) }    
    ','    { (TokComma,  _, _) }    
    not    { (TokNot,    _, _) }
    and    { (TokAnd,    _, _) }
    or     { (TokOr,     _, _) }
    impl   { (TokImpl,   _, _) }
    eqiv   { (TokEqiv,   _, _) }
    elem   { (TokElem,    _, _) }
    lbrace { (TokBraceL,  _, _) }
    rbrace { (TokBraceR,  _, _) }
    true   { (TokTru,    _, _) }
    false  { (TokFal,    _, _) }   
    var    { (TokVar $$, _, _) }

%right    equiv
%right    impl
%left     or
%left     and
%nonassoc not

%%

Expr
 : not Expr            { Pnot $2 }
 | Expr and  Expr      { Pand [$1, $3] }
 | Expr or   Expr      { Por  [$1, $3] }
 | Expr impl Expr      { Pimpl $1 $3 }
 | Expr eqiv Expr      { Pequiv $1 $3 }
 | var elem Expr1      { PvarElem (Pvar $1, $3) }
 | lpar Expr rpar      { $2 }
-- | lbrace Expr1 rbrace  { $2 }
 | true                { Ptrue }
 | false               { Pfalse }
-- | var               { Pvar $1 }
-- | selem ',' Expr    { PSelem [Pselem $1, $3] }
-- | selem	       { Pselem $1 }

Expr1
 : lbrace Expr1 rbrace { $2 }
 | selem ',' Expr1     { PSelem [Pselem $1, $3] }
 | selem	       { Pselem $1 }
  
{

--happyError _ = error ("Parse error\n")

happyError :: [Token] -> d
happyError toks = error $
                  unsafePerformIO                 
                  ( do 
                     b <- (resultOrError (test toks)) 
                     case b of 
                         Nothing -> return "unknown parse error, perhaps unexpected end of file" 
                         Just b  -> return b
                  )
   where 
     test []          = "EOF"
     test ((tok, row, col) : _)          =   ("parse error before symbol '" ++ show tok ++ "' at " ++ show (row, col)) 

resultOrError expr = do
                     result <- Control.OldException.try (Control.OldException.evaluate expr)
                     case result of 
                         Left _  -> return Nothing
                         Right b -> return (Just b)

data Pexpr name = Ptrue
                | Pfalse
                | Pvar name
                | Pselem name
                | PSelem [(Pexpr name)]                                         
                | PvarElem ((Pexpr name),(Pexpr name))
                | Pnot  (Pexpr name)
                | Pand [(Pexpr name)]
                | Por  [(Pexpr name)]
                | Pimpl  (Pexpr name) (Pexpr name)
                | Pequiv (Pexpr name) (Pexpr name)
                deriving (Ord, Eq, Show)

data LexTok = TokSElem String
            | TokAll
            | TokExists
       	    | TokParL
            | TokParR
            | TokColon 
            | TokComma
            | TokNot
            | TokTru
            | TokFal
            | TokAnd
            | TokOr
            | TokImpl
            | TokEqiv
            | TokElem
            | TokBraceL 
            | TokBraceR
            | TokVar String
            deriving Eq

simpleLexems = [(TokAll,  "ALL"), (TokExists, "EX"),
               (TokParL,   "("), (TokParR,   ")"), (TokColon,  ":"), (TokComma, ","),
               (TokNot,    "-"), (TokTru,    "1"), (TokFal,    "0"),
               (TokAnd,  "/\\"), (TokOr,   "\\/"), (TokImpl,  "=>"), (TokEqiv, "<=>"), (TokElem, "EPS"), (TokBraceL, "{"), (TokBraceR, "}") ]
 
instance Show LexTok where
  showsPrec _ (TokSElem i) = showString "SElem" . showString i
  showsPrec _ (TokVar s) = showString "VAR" . showString s
  showsPrec _ l          = case (lookup l simpleLexems) of
                Nothing -> showString "unknown lexem"
                Just s  -> showString s
         
           
type Col   = Int
type Row   = Int
type Token = (LexTok, Row, Col)

lexProp row col "" = []
lexProp row col (                 '-'  : cs)             = (TokNot,    row, col) : (lexProp row (col + 1) cs)
lexProp row col (                 '1'  : cs)             = (TokTru,    row, col) : (lexProp row (col + 1) cs)
lexProp row col (                 '0'  : cs)             = (TokFal,    row, col) : (lexProp row (col + 1) cs)
lexProp row col (                 '('  : cs)             = (TokParL,   row, col) : (lexProp row (col + 1) cs)
lexProp row col (                 ')'  : cs)             = (TokParR,   row, col) : (lexProp row (col + 1) cs)
lexProp row col (                 '{'  : cs)             = (TokBraceL, row, col) : (expandSet row (col + 1) cs)
-- lexProp row col (                       '{'  : cs)     = (TokBraceL, row, col) : (lexProp row (col + 1) cs)
-- lexProp row col (                       '}'  : cs)     = (TokBraceR, row, col) : (lexProp row (col + 1) cs)
lexProp row col (                 ':'  : cs)             = (TokColon,  row, col) : (lexProp row (col + 1) cs)
lexProp row col (                 ','  : cs)             = (TokComma,  row, col) : (lexProp row (col + 1) cs)
lexProp row col (                'E'  : 'X'  : cs)       = (TokExists, row, col) : (lexProp row (col + 2) cs)
lexProp row col (                '/'  : '\\' : cs)       = (TokAnd,    row, col) : (lexProp row (col + 2) cs)
lexProp row col (                '\\' : '/'  : cs)       = (TokOr,     row, col) : (lexProp row (col + 2) cs)
lexProp row col (                '='  : '>'  : cs)       = (TokImpl,   row, col) : (lexProp row (col + 2) cs)
lexProp row col (                'A' : 'L'  : 'L'  : cs) = (TokAll,    row, col) : (lexProp row (col + 3) cs)
lexProp row col (                '<' : '='  : '>'  : cs) = (TokEqiv,   row, col) : (lexProp row (col + 3) cs)
lexProp row col ('E' : 'L'  : 'E' : 'M' : cs)            = (TokElem,   row, col) : (lexProp row (col + 4) cs)

lexProp row col (c : cs)
  | c == '\n' = lexProp (row+1)  1        cs
  | isSpace c = lexProp  row    (col + 1) cs
  | isAlpha c = 
    let
      (ns, i, cs') = countSpan isIdChar cs
    in
      (TokVar (c:ns), row, col) : lexProp row (col + i) cs'
  | otherwise =
      error ("unexpected symbol " ++ c : " at " ++ show (row, col))
  where
    isIdChar c = isAlpha c || isDigit c || c `elem` ['\'', '_']

-- Faltet nach und nach die Menge auf und splittet in Tokens
expandSet :: Row -> Col -> [Char] -> [(LexTok, Row, Col)]
expandSet row col (c:cs)
        | c == '}'  = (TokBraceR, row, col) : (lexProp row (col + 1) cs)
        | c == ','  = (TokComma,  row, col) : (expandSet row (col + 1) cs)
        | isDigit c =
                let 
                         (ns, i, cs') = countSpan isIdChar cs
                in
                         (TokSElem ((c:ns)), row, col) : expandSet row (col + i) cs'
        | otherwise =
      error ("unexpected symbol " ++ c : " at " ++ show (row, col))
        where
    isIdChar c = isDigit c

countSpan p []         = ([], 0, [])
countSpan p xs@(x:xs')
  | p x       = (x : ys, i + 1, zs)
  | otherwise = (    [],     0, xs)
  where
    (ys, i, zs) = countSpan p xs'

parseProp = prop . lexProp 1 1

}