-----------------------------------------------------------------------------
-- |
-- Module      :  Dpfs.Simp
-- Copyright   :  (c) Thomas Ilgner 2012
-- Stability   :  provisional
-- Portability :  portable
--
-- Substitute all occurrencies of propositional variables to numbers.
-- Transforms an propositional logic formula into an conjunctive normal form.
--
-----------------------------------------------------------------------------
module Dpfs.Simp {-(numPropSubs,cnf)-} where

import Dpfs.Parser (Pexpr(..))
import Data.List
import Data.Char
import Dpfs.SimpCnf
import qualified Data.IntSet as Set

{------------------------------------------------------------------------
  Transformation aussagenlogische Formel in eine Konjunktive Normalform
-------------------------------------------------------------------------}

{- | 'cnf' transformiert eine aussagenlogische Formel in eine Konjunktive Normalform
-}
cnf fset expr  = cnfRemoveTaut  . pelimflat. pcnf . pelimflat. pcnf . pelimflat . pelimTF. pelimnot fset . pelimimpl . pelimequiv $ expr 

{- | 'cnffast' transformiert eine aussagenlogische Formel in eine Konjunktive Normalform (lineare CNF)
-}
cnffast fset expr  = cnfRemoveTaut  . pelimflat. pcnf . pelimflat. pcnf . pelimflat . pelimTF . pelimnot fset . pelimimpl .   pelimequiv . fastcnfInt $ expr

{- | 'pelimequiv' ersetzt Äquivalenzen innerhalb der Formel. (a <=> b) --> ((a => b) /\\ (b => a))
-}
pelimequiv :: Pexpr (Int, [Int]) -> Pexpr (Int, [Int])
pelimequiv Ptrue = Ptrue
pelimequiv Pfalse = Pfalse
pelimequiv (Pvar x) = (Pvar x) 
pelimequiv (Pand es) = Pand (map pelimequiv es) 
pelimequiv (Por es) =  Por (map pelimequiv es) 
pelimequiv (Pimpl e f) = Pimpl (pelimequiv e) (pelimequiv  f)  
pelimequiv (Pequiv e f) = 
    let en =  (pelimequiv e)
        fn =  (pelimequiv f)
     in Pand [Pimpl en fn, Pimpl fn en]
pelimequiv (Pnot e) =  Pnot (pelimequiv e) 

{- | 'pelimimpl' ersetzt Implikationen innerhalb der Formel. (a => b) --> (not a) \// b 
-}
pelimimpl :: Pexpr (Int, [Int]) -> Pexpr (Int, [Int])
pelimimpl Ptrue = Ptrue 
pelimimpl Pfalse = Pfalse
pelimimpl (Pvar x) = (Pvar x) 
pelimimpl (Pand es) = Pand (map pelimimpl es)  
pelimimpl (Por es) = Por (map pelimimpl es)  
pelimimpl (Pimpl e f) = Por [(Pnot (pelimimpl e)), (pelimimpl  f)]  
pelimimpl (Pnot e) =  (Pnot (pelimimpl e))

{- | 'pelimnot' ersetzt Vorkommen von (not Ptrue) bzw. (not Pfalse) zu Pfalse bzw. Ptrue
-}
pelimnot :: Int -> Pexpr (Int, [Int]) -> Pexpr (Int, [Int])
pelimnot fset Ptrue = Ptrue 
pelimnot fset Pfalse = Pfalse
pelimnot fset (Pvar x) = (Pvar x) 
pelimnot fset (Pnot (Pvar (x1,x2))) = (Pvar (x1, ([1..fset] \\ x2) )) --(Pnot (Pvar (x1,x2)))
pelimnot fset (Pnot Pfalse) = Ptrue
pelimnot fset (Pnot Ptrue) = Pfalse
pelimnot fset (Pnot (Ptrue)) = Pfalse
pelimnot fset (Pand es) = Pand (map (pelimnot fset) es)  
pelimnot fset (Por es)  = Por (map (pelimnot fset) es)  
pelimnot fset (Pnot (Pnot e)) = pelimnot fset e 
pelimnot fset (Pnot (Pand es)) = Por (map (pelimnot fset) (map Pnot es))
pelimnot fset (Pnot (Por es))  = Pand (map (pelimnot fset) (map Pnot es))

{- | 'pelimTF' löscht alle Vorkommen von Ptrue und Pfalse innerhalb der Formel.
-}
pelimTF :: Pexpr (Int, [Int]) -> Pexpr (Int, [Int])
pelimTF  xs =    
     let res = pelimTFR . pelimTFR  $ xs 
     in if res == Ptrue then Pand [Ptrue]
        else if res == Pfalse then Pand [Pfalse]
        else res

pelimTFR Ptrue = Ptrue 
pelimTFR Pfalse = Pfalse
pelimTFR (Pvar x) = (Pvar x) 
pelimTFR (Pnot x) = Pnot (pelimTFR x)
pelimTFR (Pand es) =
      let esn = map pelimTFR es
          esn2 = deleteAll Ptrue esn
       in if (Pfalse `elem` esn2) 
          then Pfalse 
          else if   (esn2 == [])
               then Ptrue 
               else Pand esn2
pelimTFR (Por es) = 
      let esn = map pelimTFR es
          esn2 = deleteAll Pfalse esn
       in if (Ptrue `elem` esn2) 
          then Ptrue 
          else if  (esn2 == [])
               then Pfalse 
               else Por esn2

deleteAll x [] = []
deleteAll x (y:ys) = if x == y then deleteAll x ys else y : (deleteAll x ys)

{- | 'pelimflat' fasst verschachtelte Und- und Oder-Verknüpfungen zusammen.
-}
pelimflat :: Pexpr (Int, [Int]) -> Pexpr (Int, [Int])
pelimflat x =
   let xn = pelimflatr x
   in if x /= xn then pelimflat xn
      else xn

pelimflatr (Por es) = if (any  portop es) 
       then Por (concat (map porsublst es)) 
       else (Por (map pelimflatr es))
pelimflatr (Pand es) = if (any  pandtop es) 
       then Pand (concat (map pandsublst es)) 
       else (Pand (map pelimflatr es))
pelimflatr x = x
 
porsublst (Por e) = e
porsublst  e = [e]

pandsublst (Pand e) = e
pandsublst  e = [e] 

portop (Por _) = True
portop _ = False
pandtop (Pand _) = True
pandtop _ = False

{- | 'cnfRemoveTaut' löscht Klauseln:
-- indem Atome als Menge genau die festgelegte Menge besitzen, und 
-- indem ein Atom x mehrfach vorkommt und die Mengen dieser Atome disjunkt sind.
-}
cnfRemoveTaut :: Pexpr (Int, [Int]) -> Pexpr (Int, [Int])
cnfRemoveTaut (Pand cls) =  Pand (cnfRemoveTaut' cls)
cnfRemoveTaut cls 	 =  error ("keine Klauselmenge " ++ show cls)

cnfRemoveTaut'  [] = []
cnfRemoveTaut'  ((Por cl):clr)
  | (cl ==  [Ptrue]) || cnfRemoveTaut'' cl = cnfRemoveTaut' clr
  | otherwise = (Por cl): (cnfRemoveTaut' clr)

cnfRemoveTaut'' [] = False
cnfRemoveTaut'' cl@((Pvar (x1,x2)):clr)
  | not $ null $ filter (findTaut (x1,x2)) clr = True
  | otherwise = cnfRemoveTaut'' clr
  where
    findTaut (x1,x2) (Pvar (y1,y2))
      | (x1 == y1) && ((null $ x2 `intersect` y2)) = True
      | otherwise  = False
      
{--------------------------------------------------------------------
  Variablenersetzung
--------------------------------------------------------------------}

{- | 'numPropSubs' substituiert alle vorkommenden Variablen durch Int-Zahlen und zusätzlich wird der Typ PvarElem ersetzt durch Pvar. Das Ergebnis ist ein Tupel aus dem Variablen-Bindungen zu den substituierten Zahlen und der substituierte Liste selbst .
-}  
numPropSubs :: Int -> Pexpr String -> ([(String, Int)], Pexpr (Int, [Int]))
numPropSubs fset x = let  ((_,asso),p) = pVarNumber fset (1, []) x
		     in (asso,p)

{- | 'pVarNumber' substituiert alle vorkommenden Variablen durch Int-Zahlen und ersetzt den Typ PvarElem durch Pvar (Num,Liste der Elemente). 
-}
pVarNumber :: Int -> (Int, [(String, Int)]) -> Pexpr String -> ((Int, [(String, Int)]), Pexpr (Int, [Int]))
pVarNumber fset bnds  Ptrue   = (bnds, Ptrue)
pVarNumber fset bnds  Pfalse  = (bnds, Pfalse)
pVarNumber fset bnds (Pnot e) = (bnds', Pnot e') where (bnds', e') = pVarNumber fset bnds e
pVarNumber fset bnds (PvarElem (e1,e2)) = 
        let s     = showVar e1
            s1    = pSelemsToList e2
            true  = [1..fset] == s1
            false = null s1
        in
        if true then
            (bnds, Ptrue)
        else if false then
            (bnds, Pfalse)
            else       
                case (lookup s table) of
                    Nothing     -> ((i+1, ((s),i) : table), (Pvar ( i,s1 )) )
                    Just n      -> (bnds,                   (Pvar ( n,s1 )) )
        where
            (i, table) = bnds
pVarNumber fset bnds (Pand es)      = pVarNumberList fset bnds Pand es
pVarNumber fset bnds (Por  es)      = pVarNumberList fset bnds Por  es
pVarNumber fset bnds (Pimpl  e1 e2) = pVarNumberList fset bnds (foldr1 Pimpl)  [e1, e2]
pVarNumber fset bnds (Pequiv e1 e2) = pVarNumberList fset bnds (foldr1 Pequiv) [e1, e2]		     


{--------------------------------------------------------------------
  Helpers
--------------------------------------------------------------------}

pVarNumberList :: Int -> (Int, [(String, Int)]) -> ([Pexpr (Int, [Int])] -> Pexpr (Int, [Int])) -> [Pexpr String] -> ((Int, [(String, Int)]), Pexpr (Int, [Int]))
pVarNumberList fset bnds c es =
  (bnds', c es') where (bnds', es') = mapAccumL (pVarNumber fset) bnds es

pSelemsToList :: Pexpr String -> [Int]
pSelemsToList (PSelem (x:xs)) = pSelemsToList' (x:xs)
pSelemsToList (Pselem x) = [read x::Int]

pSelemsToList' [] = []                
pSelemsToList' ((Pselem x):xs) = (read x::Int) : pSelemsToList' xs
pSelemsToList' xs = pSelemsToList (head xs)

showVar (Pvar x) = x
